import unittest
from pascal_voc_tools.tools import bb_intersection_over_union


class TestTools(unittest.TestCase):
    def test_bb_intersection_over_union(self):
        bboxA = [0, 0, 10, 10]
        bboxB = [0, 0, 10, 10]
        iou = bb_intersection_over_union(bboxA, bboxB)
        self.assertEqual(iou, 1.0)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import logging
from collections import OrderedDict

logger = logging.getLogger(__name__)


class Layer():
    """Darknet Layer Struct"""
    def __init__(self, name):
        """
        Arguments:
            name: str, layer name.
        """
        self.name = name
        self.param = OrderedDict()


class DarknetConfig():
    """Tools about config file for darknet.
    """
    def __init__(self):
        self.layers = None

    def parse(self, config_path):
        """decode the config file for darknet
        Arguments:
            config_path: str, the path of config file;
        Returns:
            layers: list, including Layer struct.
        """
        with open(config_path) as f:
            config_data = f.read().strip().split('\n')
            # config_data = [line.strip() for line in config_data if line]
            # config_data = [line for line in config_data if line[0] != '#']
        layers = []
        layers_str = ('\n'.join(config_data)).split('[')
        for layer_str in layers_str:
            if not layer_str:
                continue
            name, param_str = layer_str.split(']')
            param_list = param_str.strip().split('\n')
            new_layer = Layer(name)
            for param in param_list:
                if param.strip() == '' or param[0] == '#':
                    key = param
                    value = ''
                else:
                    key, value = param.strip().split('=')
                    key = key.strip()
                    if key in new_layer.param:
                        logger.warning(
                            f'The key {key} has been appended, ' +
                            f'the value is {new_layer.param[key]}'
                            )
                new_layer.param[key] = value.strip()
            layers.append(new_layer)
        self.layers = layers
        return layers

    def change_anchors(self, layers, new_anchors):
        """
        Arguments:
            layers: list, the list of Layers;
            new_anchors: list, which have n anchors, for each
                         anchor it has the format like [w, h].
        Returns:
            layers: list, the layers had been changed.
        """
        for i, layer in enumerate(layers):
            if layer.name == 'yolo':
                mask = layer.param['mask'].split(',')
                mask = [int(s.strip()) for s in mask]
                anchors_str = []
                for mask_index in mask:
                    assert mask_index < len(new_anchors)
                anchors_str = [
                    '{:.2f},{:.2f}'.format(anchor[0], anchor[1])
                    for anchor in new_anchors
                ]
                layers[i].param['anchors'] = ', '.join(anchors_str)
                layers[i].param['num'] = str(len(new_anchors))
        return layers

    def save(self, layers, save_path):
        """Saving the list of Layer to save_path.
        Arguments:
            layers: list, have some Layers;
            save_path: str, the path using to save.
        """
        config_data = []
        for layer in layers:
            param_list = []
            for key, val in layer.param.items():
                if key.strip() == '' or key[0] == '#':
                    param_list.append((key + ' ' + val).strip())
                else:
                    param_list.append(key + ' = ' + val)
            name = '[' + layer.name + ']'
            config_data.append(name + '\n' + '\n'.join(param_list) + '\n')
        config_data = '\n'.join(config_data)

        with open(save_path, 'w') as f:
            f.write(config_data)
        print('Data have been saved in {}.'.format(save_path))
        return 0

    def check_yolo(self, cfg_file, label_list_file):
        """Using to check some key in yolo config.
        Arguments:
            cfg_file: str, the config file path;
            label_list_file: str, the file of label list;
        """
        assert os.path.exists(cfg_file), cfg_file
        assert os.path.exists(label_list_file), label_list_file

        # get labels
        with open(label_list_file) as f:
            labels = f.read().strip().split('\n')

        layers = self.parse(cfg_file)
        for i in range(len(layers) - 1, 1):
            if layers[i].name == 'yolo':
                # check classes
                assert int(layers[i].param['classes']) == len(labels)
                # check anchors number
                anchors = layers[i].param['anchors'].split(',')
                assert anchors % 2 == 0, f'anchors length is wrong: {anchors}'
                # check num
                assert int(layers[i].param['num']) == len(anchors / 2)
                # check mask
                mask = list(map(int, layers[i].param['mask'].split(',')))
                for m in mask:
                    assert m < anchors / 2, 'mask is not right: {}'.format(
                        mask)
                # check filters
                assert layers[i - 1].name == 'convolutional'
                assert int(
                    layers[i - 1].param['filters']) == (len(labels) + 5) * len(
                        mask), 'filters is not right: {}'.format(
                            layers[i - 1].param['filters'])
        print('Check Done.')
        return 0


def main():
    from kmeans import KMeansAnchors
    cfg_file = './configs/yolov3-20190423.cfg'
    config = DarknetConfig()
    layers = config.parse('./configs/yolov3-20190423.cfg-backup')
    width = layers[0].param['width']
    yolo_input_shape = int(width)
    kmeans = KMeansAnchors()
    anchors = kmeans.calculate(yolo_input_shape=yolo_input_shape)
    print("Result: {}".format(anchors))
    layers = config.change_anchors(layers, anchors)
    config.save(layers, cfg_file)


if __name__ == '__main__':
    main()



# -*- coding:utf-8 -*-
"""
decode xml file of Pascal Voc annotation write a xml file
about pascal voc annotation.
"""
import os
import logging

from lxml import etree
from lxml.etree import SubElement

from .tools import bb_intersection_over_union as iou

logger = logging.getLogger(__name__)


def get_first_node_info(node, key, default=''):
    val = node.xpath(f'./{key}/text()')
    if val:
        return val[0]
    else:
        return default


def set_sub_node_info(node, key, val):
    sub_node = SubElement(node, key)
    sub_node.text = str(val)
    return node


class ImageSize(object):
    """Size data format in Pascal VOC xml

    Attributes:
        width: a int of image width.
        height: a int of image height.
        depth: a int of image depth.
    """
    def __init__(self, width: int = 0, height: int = 0, depth: int = 0):
        self.width = width
        self.height = height
        self.depth = depth

    def __str__(self):
        return f"ImageSize({self.width}, {self.height}, {self.depth})"

    def load_from_node(self, node):
        """
        """
        self.width = int(get_first_node_info(node, 'width', 0))
        self.height = int(get_first_node_info(node, 'height', 0))
        self.depth = int(get_first_node_info(node, 'depth', 0))
        return self

    def save_to_node(self, node):
        """
        """
        set_sub_node_info(node, 'width', self.width)
        set_sub_node_info(node, 'height', self.height)
        set_sub_node_info(node, 'depth', self.depth)
        return self


class DataSource(object):
    """Source data format in Pascal VOC xml

    Attributes:
        database: a str of dataset name.
    """
    def __init__(self, database: str = 'Unknown'):
        self.database = database

    def __str__(self):
        return f"DataScorce({self.database})"

    def load_from_node(self, node):
        """
        """
        self.database = get_first_node_info(node, 'database', 'Unknown')
        return self

    def save_to_node(self, node):
        """
        """
        set_sub_node_info(node, 'database', self.database)
        return self


class Bndbox(object):
    """Bndbox data format in Object for Pascal VOC xml

    Attributes:
        xmin: a int of bndbox left.
        ymin: a int of bndbox top.
        xmax: a int of bndbox right.
        ymax: a int of bndbox bottom.
    """
    def __init__(self,
                 xmin: int = 0,
                 ymin: int = 0,
                 xmax: int = 0,
                 ymax: int = 0):
        self.xmin = xmin
        self.ymin = ymin
        self.xmax = xmax
        self.ymax = ymax

    def __str__(self):
        return f"Bndbox({self.xmin}, {self.ymin}, {self.xmax}, {self.ymax})"

    def load_from_node(self, node):
        """decode bndbox node info

        Arguments:
            node: a bndbox node in xml.
        """
        self.xmin = int(get_first_node_info(node, 'xmin', 0))
        self.ymin = int(get_first_node_info(node, 'ymin', 0))
        self.xmax = int(get_first_node_info(node, 'xmax', 0))
        self.ymax = int(get_first_node_info(node, 'ymax', 0))
        return self

    def save_to_node(self, node):
        """encode xmin, ymin, xmax, ymax to bndbox node.

        Arguments:
            node: a bndbox SubElement node.
        """
        set_sub_node_info(node, 'xmin', self.xmin)
        set_sub_node_info(node, 'ymin', self.ymin)
        set_sub_node_info(node, 'xmax', self.xmax)
        set_sub_node_info(node, 'ymax', self.ymax)
        return self

    def convert2relative_xywh(self, size):
        """from absolute coordinate to relative coordinate

        Arguments:
            size: a tuple of width and height
        """
        box = (float(self.xmin), float(self.xmax), float(self.ymin),
               float(self.ymax))
        dw = 1. / (size[0])
        dh = 1. / (size[1])
        x = max((box[0] + box[1]) / 2.0 - 1, 0)
        y = max((box[2] + box[3]) / 2.0 - 1, 0)
        w = box[1] - box[0]
        h = box[3] - box[2]
        x = x * dw
        w = w * dw
        y = y * dh
        h = h * dh
        return (x, y, w, h)

    def resize(self, rate: float, horizion_bias=0, vertical_bias=0):
        """resize a bbox

        Args:
            rate: box change rate;
            horizion_bias: xmin and xmax will add it;
            vertical_bias: ymin and ymax will add it,
        Returns:
            new bbox list
        """
        rate = float(rate)

        self.xmin = int(self.xmin * rate) + horizion_bias
        self.ymin = int(self.ymin * rate) + vertical_bias
        self.xmax = int(self.xmax * rate) + horizion_bias
        self.ymax = int(self.ymax * rate) + vertical_bias
        return self


class XmlObject(object):
    """Object data foramt in Pascal VOC xml

    Attributes:
        name: the category name.
        pose: object post description.
        truncated: default is 0.
        difficult: default is 0.
        bndbox: the Bndbox.
    """
    def __init__(self,
                 name: str = '',
                 pose: str = 'Unspecified',
                 truncated: int = 0,
                 difficult: int = 0,
                 bndbox: Bndbox = Bndbox()):
        self.name = name
        self.pose = pose
        self.truncated = truncated
        self.difficult = difficult
        self.bndbox = bndbox

    def __str__(self):
        return f"XmlObject({self.name}, {self.pose}," + \
            f" {self.truncated}, {self.difficult}, {self.bndbox})"

    def load_from_node(self, node):
        """
        """
        self.name = get_first_node_info(node, 'name', '')
        self.pose = get_first_node_info(node, 'pose', 'Unspecified')
        self.truncated = int(get_first_node_info(node, 'truncated', 0))
        self.difficult = int(get_first_node_info(node, 'difficult', 0))
        node_bndbox = node.xpath('./bndbox')
        if node_bndbox:
            self.bndbox = Bndbox().load_from_node(node_bndbox[0])
        return self

    def save_to_node(self, node):
        """
        """
        set_sub_node_info(node, 'name', self.name)
        set_sub_node_info(node, 'pose', self.pose)
        set_sub_node_info(node, 'truncated', self.truncated)
        set_sub_node_info(node, 'difficult', self.difficult)

        node_bndbox = SubElement(node, 'bndbox')
        self.bndbox.save_to_node(node_bndbox)
        return self


class PascalXml(object):
    """Pascal VOC xml file data format

    Attributes:
        folder: image folder.
        filename: image file name.
        path: image file path.
        source: the DataSource.
        size: the ImageSize.
        segmented: default 0.
        object_list: a list of XmlObject
    """
    def __init__(self,
                 folder: str = '',
                 filename: str = '',
                 path: str = '',
                 source: DataSource = DataSource(),
                 size: ImageSize = ImageSize(),
                 segmented: int = 0,
                 object_list: list = []):
        self.folder = folder
        self.filename = filename
        self.path = path if path else os.path.join(folder, filename)
        self.source = source
        self.size = size
        self.segmented = segmented
        self.object = object_list

    def load(self, xml_file_path):
        """form a xml file load data.

        Arguments:
            xml_file_path: a xml file path.
        """
        # load_pascal_xml(xml_file_path, self)
        html = etree.parse(xml_file_path)
        annotation = html.xpath('/annotation')  # load first annotation
        if not annotation:
            logger.error("Can not find annotation node")
            raise KeyError("Can not find annotation node")

        self.load_from_node(annotation[0])

        return self

    def load_from_node(self, node):
        """load data from Element

        Arguments:
            node: an annotation Element.
        """
        self.folder = get_first_node_info(node, 'folder', './')
        self.filename = get_first_node_info(node, 'filename', '')
        self.path = get_first_node_info(
            node, 'path', os.path.join(self.folder, self.filename))

        node_source = node.xpath('./source')
        if node_source:
            self.source = DataSource().load_from_node(node_source[0])

        node_size = node.xpath('./size')
        if node_size:
            self.size = ImageSize().load_from_node(node_size[0])
        self.segmented = int(get_first_node_info(node, 'segmented', 0))

        self.object = []
        for obj in node.xpath('./object'):
            xml_obj = XmlObject().load_from_node(obj)
            self.object.append(xml_obj)
        return self

    def save(self, save_xml_path):
        """save data to a xml file.

        Arguments:
           save_xml_path: the xml path to save data.
        """
        # save_pascal_xml(save_xml_path, self)
        node_root = etree.Element('annotation')
        self.save_to_node(node_root)
        tree = etree.ElementTree(node_root)
        tree.write(save_xml_path,
                   pretty_print=True,
                   xml_declaration=False,
                   encoding='utf-8')
        return self

    def save_to_node(self, node):
        """save info to annotation Element

        Arguments:
            node: an annotation Element.
        """
        set_sub_node_info(node, 'folder', self.folder)
        set_sub_node_info(node, 'filename', self.filename)
        set_sub_node_info(node, 'path', self.path)

        node_source = SubElement(node, 'source')
        self.source.save_to_node(node_source)

        set_sub_node_info(node, 'segmented', self.segmented)

        node_size = SubElement(node, 'size')
        self.size.save_to_node(node_size)

        for obj in self.object:
            node_object = SubElement(node, 'object')
            obj.save_to_node(node_object)
        return self

    def replace_name(self, old_name, new_name):
        """Replace an object name.

        Args:
            old_name: str, an object class name.
            new_name: str, a new object class name.
        """
        for obj in self.object:
            if obj.name == old_name:
                obj.name = new_name
        return self

    def convert2yolotxt(self, save_path: str, classes: list):
        """save data to txt for yolo format.

        Arguments:
            save_path: the txt file path to save data.
            classes: a list of categories.
        """
        assert save_path[
            -4:] == '.txt', f"Please check save_path is right: {save_path}"

        out_file = open(save_path, 'w')

        w = int(self.size.width)
        h = int(self.size.height)

        for obj in self.object:
            difficult = obj.difficult
            cls = obj.name
            if cls not in classes or int(difficult) == 1:
                continue
            cls_id = classes.index(cls)
            bb = obj.bndbox.convert2relative_xywh((w, h))
            out_file.write(
                str(cls_id) + " " + " ".join([str(a) for a in bb]) + '\n')

        out_file.close()

        return self

    def convert2csv(self, classes: list = []):
        """Convert data to info list. The head is
        [file_name, width, height, category, xmin, ymin, xmax, ymax]

        Arguments:
            classes: a list of categories.
        """
        file_name = self.filename
        width = self.size.width
        height = self.size.height

        info_list = []
        for bbox in self.object:
            category = bbox.name
            xmin = bbox.bndbox.xmin
            ymin = bbox.bndbox.ymin
            xmax = bbox.bndbox.xmax
            ymax = bbox.bndbox.ymax

            if classes and category not in classes:
                continue
            info_list.append(
                [file_name, width, height, category, xmin, ymin, xmax, ymax])
        return info_list

    def add_object(self,
                   name,
                   xmin,
                   ymin,
                   xmax,
                   ymax,
                   pose='Unspecified',
                   truncated=0,
                   difficult=0):
        """add an object info

        Args:
            name: str, class name.
            xmin: int, left.
            ymin: int, top.
            xmax: int, right.
            ymax: int, bottom.
            pose: str, default is 'Unspecified'.
            truncated: str, default is 0.
            difficult: int, default is 0.
        """
        bndbox = Bndbox(xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)
        obj = XmlObject(name=name,
                        pose=pose,
                        truncated=truncated,
                        difficult=difficult,
                        bndbox=bndbox)
        self.object.append(obj)
        return self

    def resize_obj_by_rate(self, rate: float, biases: tuple):
        """Resize all bndbox by rate.

        Arguments:
            rate: a float for timeing bndbox.
        """
        original_width = int(self.size.width)
        original_height = int(self.size.height)
        new_width = int(original_width * rate)
        new_height = int(original_height * rate)

        # bias
        vertical_bias, horizion_bias = biases

        self.size.width = new_width
        self.size.height = new_height

        for obj in self.object:
            obj.bndbox = obj.bndbox.resize(rate, horizion_bias, vertical_bias)

        return self

    def crop_annotations(self, split_bboxes, iou_thresh=0.7):
        """Using split_bboxes to split an xml file.
        Arguments:
            xml_info: dict, all info about a xml.
            split_bboxes: list, like [[xmin, ymin, xmax, ymax], ]
        Returns:
            subannotations: list, like [xml_info, ]
        """
        subannotations = []
        for image_bbox in split_bboxes:
            img_xmin, img_ymin, img_xmax, img_ymax = image_bbox

            # init sub xml info
            sub_xml = PascalXml()
            sub_xml.folder = self.folder
            sub_xml.path = self.path
            sub_xml.filename = self.filename
            sub_xml.size = ImageSize(width=img_xmax - img_xmin,
                                     height=img_ymax - img_ymin,
                                     depth=self.size.depth)
            sub_xml.source = DataSource(self.source.database)
            sub_xml.segmented = self.segmented
            sub_xml.object = []
            for bbox in self.object:
                ob_xmin = bbox.bndbox.xmin
                ob_ymin = bbox.bndbox.ymin
                ob_xmax = bbox.bndbox.xmax
                ob_ymax = bbox.bndbox.ymax

                input_xmin = min(img_xmax, max(ob_xmin, img_xmin))
                input_ymin = min(img_ymax, max(ob_ymin, img_ymin))
                input_xmax = max(img_xmin, min(ob_xmax, img_xmax))
                input_ymax = max(img_ymin, min(ob_ymax, img_ymax))
                if iou([ob_xmin, ob_ymin, ob_xmax, ob_ymax],
                       [input_xmin, input_ymin, input_xmax, input_ymax
                        ]) > iou_thresh:
                    sub_bbox = Bndbox(input_xmin - img_xmin,
                                      input_ymin - img_ymin,
                                      input_xmax - img_xmin,
                                      input_ymax - img_ymin)
                    sub_obj = XmlObject(name=bbox.name,
                                        bndbox=sub_bbox,
                                        truncated=bbox.truncated,
                                        difficult=bbox.difficult)
                    sub_xml.object.append(sub_obj)
            subannotations.append(sub_xml)

        return subannotations


def load_pascal_xml(xml_file_path: str, default_format=None) -> PascalXml:
    """Load a pascal format xml file.
    Arguments:
        xml_file_path: a xml file path.
        default_format: PascalXml instance.
    Returns:
        default_format: the PascalXml instance including data.
    Raises:
        ListError: can not find the key in xml file.
    """
    if not default_format:
        default_format = PascalXml()

    html = etree.parse(xml_file_path)
    annotation = html.xpath('/annotation')  # load first annotation
    if not annotation:
        logger.error("Can not find annotation node")
        raise KeyError("Can not find annotation node")

    # for xml only one annotation node
    for ann in annotation:
        folder = get_first_node_info(ann, 'folder', './')
        filename = get_first_node_info(ann, 'filename', '')
        path = get_first_node_info(ann, 'path', os.path.join(folder, filename))
        database = get_first_node_info(ann, 'source/database', 'Unknown')
        width = get_first_node_info(ann, 'size/width', 0)
        height = get_first_node_info(ann, 'size/height', 0)
        depth = get_first_node_info(ann, 'size/depth', 0)
        segmented = get_first_node_info(ann, 'segmented', 0)

    default_format.folder = folder
    default_format.filename = filename
    default_format.path = path
    default_format.source = DataSource()
    default_format.source.database = database
    default_format.size = ImageSize(int(width), int(height), int(depth))
    default_format.segmented = int(segmented)

    default_format.object = []
    for obj in ann.xpath('./object'):
        name = get_first_node_info(obj, 'name')
        pose = get_first_node_info(obj, 'pose')
        truncated = get_first_node_info(obj, 'truncated', 0)
        difficult = get_first_node_info(obj, 'difficult', 0)
        xmin = get_first_node_info(obj, 'bndbox/xmin', 0)
        ymin = get_first_node_info(obj, 'bndbox/ymin', 0)
        xmax = get_first_node_info(obj, 'bndbox/xmax', 0)
        ymax = get_first_node_info(obj, 'bndbox/ymax', 0)

        bndbox = Bndbox(int(xmin), int(ymin), int(xmax), int(ymax))
        xml_obj = XmlObject(name=name,
                            pose=pose,
                            truncated=int(truncated),
                            difficult=int(difficult),
                            bndbox=bndbox)
        default_format.object.append(xml_obj)

    return default_format


def save_pascal_xml(save_xml_path: str, pascal_xml: PascalXml) -> PascalXml:
    """save data to xml
    Arguments:
        save_xml_path: save path.
        pascal_xml: PascalXml.
    Returns:
        PascalXml
    """
    node_root = etree.Element('annotation')

    set_sub_node_info(node_root, 'folder', pascal_xml.folder)
    set_sub_node_info(node_root, 'filename', pascal_xml.filename)
    set_sub_node_info(node_root, 'path', pascal_xml.path)

    node_source = SubElement(node_root, 'source')
    set_sub_node_info(node_source, 'database', pascal_xml.source.database)

    set_sub_node_info(node_root, 'segmented', pascal_xml.segmented)

    node_size = SubElement(node_root, 'size')
    set_sub_node_info(node_size, 'width', pascal_xml.size.width)
    set_sub_node_info(node_size, 'height', pascal_xml.size.height)
    set_sub_node_info(node_size, 'depth', pascal_xml.size.depth)

    for obj in pascal_xml.object:
        node_object = SubElement(node_root, 'object')
        set_sub_node_info(node_object, 'name', obj.name)
        set_sub_node_info(node_object, 'truncated', obj.truncated)
        set_sub_node_info(node_object, 'difficult', obj.difficult)
        set_sub_node_info(node_object, 'name', obj.name)
        set_sub_node_info(node_object, 'name', obj.name)

        node_bndbox = SubElement(node_object, 'bndbox')
        set_sub_node_info(node_bndbox, 'xmin', obj.bndbox.xmin)
        set_sub_node_info(node_bndbox, 'ymin', obj.bndbox.ymin)
        set_sub_node_info(node_bndbox, 'xmax', obj.bndbox.xmax)
        set_sub_node_info(node_bndbox, 'ymax', obj.bndbox.ymax)

    tree = etree.ElementTree(node_root)
    tree.write(save_xml_path,
               pretty_print=True,
               xml_declaration=False,
               encoding='utf-8')

    return pascal_xml


class XmlParser():
    """Code and Decode the xml file"""
    def __init__(self):
        raise ("Please use new class PascalXml")



# coding:utf-8

import os
import glob
import logging

logger = logging.getLogger(__name__)


class JPEGImages(object):
    """Some tools for JPEGImages dir.
    """
    def __init__(self, image_dir=None):
        self.dir = image_dir
        self.jpg_list = []

    def load(self, image_dir=None):
        """Load jpg image files' path.

        Arguments:
            image_dir: a directory incluing jpg images.
        """
        if image_dir:
            self.dir = image_dir

        assert os.path.isdir(self.dir), self.dir
        self.jpg_list = sorted(glob.glob(os.path.join(self.dir, '*.jpg')))

        logger.info(f'Load {len(self.jpg_list)} images.')
        return self



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import logging

logger = logging.getLogger(__name__)


def bb_intersection_over_union(boxA, boxB):
    """calculate intersection over union between two boundboxes.

    Args:
        boxA: list of xmin, ymin, xmax, ymax;
        boxB: list of xmin, ymin, xmax, ymax;
    Returns:
        a float number of iou between two inputs.
    """
    # determine the (x, y)-coordinates of the intersection rectangle
    xA = max(boxA[0], boxB[0])
    yA = max(boxA[1], boxB[1])
    xB = min(boxA[2], boxB[2])
    yB = min(boxA[3], boxB[3])

    # compute the area of intersection rectangle
    interArea = max(0, xB - xA + 1) * max(0, yB - yA + 1)

    # compute the area of both the prediction and ground-truth
    # rectangles
    boxAArea = (boxA[2] - boxA[0] + 1) * (boxA[3] - boxA[1] + 1)
    boxBArea = (boxB[2] - boxB[0] + 1) * (boxB[3] - boxB[1] + 1)

    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the interesection area
    try:
        iou = interArea / float(boxAArea + boxBArea - interArea)
    except ZeroDivisionError as err:
        logger.exception(err)
        print(boxA, boxB)
    # return the intersection over union value
    return iou


def average_precision_11_point():
    return


def average_precision_matric():
    return



# -*- coding:utf-8 -*-
"""
Some tools for VOC data set dir.
"""
import glob
import logging
import os

import tqdm

from .annotations_tools import Annotations
from .image_tools import ImageWrapper
from .jpegimages_tools import JPEGImages
from .xml_tools import PascalXml

logger = logging.getLogger(__name__)


class Main(object):
    def __init__(self, main_dir):
        self.dir = main_dir
        self.name_list_map = {}

    def __str__(self):
        return f"Main(main_dir={self.dir})"

    def load(self, set_name):
        file_path = os.path.join(self.dir, set_name + '.txt')
        assert os.path.isfile(file_path), file_path

        with open(file_path) as f:
            name_lines = f.read().strip().split('\n')
        logger.info(f'There are {len(name_lines)} names in the file.')

        self.name_list_map[set_name] = name_lines

        return self

    def save(self):
        for set_name in self.name_list_map:
            save_path = os.path.join(self.dir, set_name + '.txt')
            with open(save_path, 'w') as f:
                f.write('\n'.join(self.name_list_map[set_name]))
        return self


class VOCTools(object):
    """a VOC format dataset.

    Attributes:
        root_dir: a voc directory like VOC2007.
    """
    def __init__(self, voc_root_dir: str):
        self.root_dir = voc_root_dir
        self.year = self.get_year()

        self.annotations = Annotations(
            os.path.join(self.root_dir, 'Annotations'))
        self.jpegimages = JPEGImages(os.path.join(self.root_dir, 'JPEGImages'))
        self.main = Main(os.path.join(self.root_dir, 'ImageSets/Main'))

    def __str__(self):
        return f"VOCTools(voc_root_dir={self.root_dir})"

    def get_year(self):
        year = None
        sub_dir = os.path.basename(self.root_dir.rstrip('/'))
        if sub_dir and 'VOC' == sub_dir[:3]:
            year = sub_dir[3:]
        return year

    def gen_format_dir(self):
        ann_dir = self.annotations.dir
        jpg_dir = self.jpegimages.dir
        main_dir = self.main.dir
        if not os.path.exists(ann_dir):
            os.makedirs(ann_dir)
        if not os.path.exists(jpg_dir):
            os.makedirs(jpg_dir)
        if not os.path.exists(main_dir):
            os.makedirs(main_dir)
        return self

    def resize_by_size(self, width, height, save_root_dir):
        """Resize whole dataset by set a fix image size.

        Args:
            width: int, new image width.
            height: int, new image height.
            save_root_dir: a new dir to save data.

        Returns:
            new VOCTools have new data.
        """
        self.annotations.load()
        save_voc = VOCTools(save_root_dir)
        save_voc.gen_format_dir()

        logger.info('Resizing dataset ...')
        for xml_path in tqdm.tqdm(self.annotations.ann_list):
            image_path = self.get_image_path_by_xml_path(xml_path)
            xml_name = os.path.basename(xml_path)
            save_xml_path = os.path.join(save_voc.annotations.dir, xml_name)
            image_name = os.path.basename(image_path)
            save_image_path = os.path.join(save_voc.jpegimages.dir, image_name)

            # resize image
            image = ImageWrapper().load(image_path)
            rate, biases = image.resize_letter_box(width, height)
            image.save(save_image_path)

            # resize annotation and save
            xml_data = PascalXml().load(xml_path).resize_obj_by_rate(
                rate, biases)
            # rewrite info
            xml_data.foler = save_voc.jpegimages.dir
            xml_data.filename = image_name
            xml_data.path = save_image_path
            # letter box resize will pad zero
            xml_data.size.width = width
            xml_data.size.height = height
            xml_data.save(save_xml_path)

        return save_voc

    def get_image_path_by_xml_path(self, xml_path):
        """using xml path to inference image path

        Args:
            xml_path: str, xml file path.

        Returns:
            image path in current dataset.
        """
        xml_name = os.path.basename(xml_path)
        image_path = os.path.join(self.jpegimages.dir,
                                  xml_name.replace('.xml', '.jpg'))
        return image_path

    def match_jpg_xml(self):
        self.jpegimages.load(self.jpegimages.dir)
        self.annotations.load()

        images_name = [
            os.path.basename(path).split('.')[0]
            for path in self.jpegimages.jpg_list
        ]
        xmls_name = [
            os.path.basename(path).split('.')[0]
            for path in self.annotations.ann_list
        ]

        matched_name = [name for name in xmls_name if name in images_name]

        matched_images_path = [
            os.path.join(self.jpg_dir, name + '.jpg') for name in matched_name
        ]
        matched_xmls_path = [
            os.path.join(self.xml_dir, name + '.xml') for name in matched_name
        ]
        return matched_images_path, matched_xmls_path

    def get_crop_bboxes(self,
                        width,
                        height,
                        min_side,
                        max_side,
                        cover_thresh=0.2):
        bboxes = []
        if width >= height:
            x_stride = max_side
            y_stride = min_side
        else:
            x_stride = min_side
            y_stride = max_side

        def add_row_crop(start_y):
            start_x = 0
            while start_x + x_stride < width:
                bboxes.append(
                    [start_x, start_y, start_x + x_stride, start_y + y_stride])
                start_x += int(x_stride * (1 - cover_thresh))
            bboxes.append([
                max(0, width - x_stride), start_y, width,
                min(height, start_y + y_stride)
            ])

        start_y = 0
        while start_y + y_stride < height:
            add_row_crop(start_y)
            start_y += int(y_stride * (1 - cover_thresh))
        add_row_crop(max(0, height - y_stride))
        return bboxes

    def crop_data(self,
                  save_root_dir,
                  set_name_list,
                  min_side=1000,
                  max_side=1333,
                  cover_thresh=0.2,
                  iou_thresh=0.7):
        """split image and annotation to some sub data.

        Arguments:
            save_root_dir: a new dir to save new data.
            set_name_list: a list like [train, val].
            min_side: a int of sub image min side.
            max_side: a int of sub image max side.
            cover_thresh: sub image overlap rate.
            iou_thresh: bndbox with sub image overlap rate.

        Returns:
            a new VOCTools including new data.
        """
        new_voc = VOCTools(save_root_dir)
        new_voc.gen_format_dir()

        for set_name in set_name_list:
            self.main.load(set_name)

        new_voc.main.name_list_map = {}
        for set_name in self.main.name_list_map:
            name_list = self.main.name_list_map[set_name]
            logger.info(f'Start crop from {set_name} set')
            new_voc.main.name_list_map[set_name] = []

            for name_id in tqdm.tqdm(name_list):
                xml_path = os.path.join(self.annotations.dir, name_id + '.xml')
                jpg_path = os.path.join(self.jpegimages.dir, name_id + '.jpg')

                images, xml_writers = self.crop_image_annotations(
                    jpg_path,
                    xml_path,
                    min_side,
                    max_side,
                    cover_thresh=cover_thresh,
                    iou_thresh=iou_thresh)
                for i, (image,
                        xml_writer) in enumerate(zip(images, xml_writers)):
                    new_name_id = name_id + '_{:0>2d}'.format(i)
                    new_voc.main.name_list_map[set_name].append(new_name_id)
                    save_xml_path = os.path.join(new_voc.annotations.dir,
                                                 new_name_id + '.xml')
                    save_jpg_path = os.path.join(new_voc.jpegimages.dir,
                                                 new_name_id + '.jpg')
                    image.save(save_jpg_path)
                    xml_writer.folder = new_voc.jpegimages.dir
                    xml_writer.path = save_jpg_path
                    xml_writer.save(save_xml_path)

        new_voc.main.save()
        return new_voc

    def crop_image_annotations(self,
                               jpg_path,
                               xml_path,
                               min_side,
                               max_side,
                               cover_thresh=0.2,
                               iou_thresh=0.7):
        """Split an image and it's annotation file.

        Arguments:
            jpg_path: str, image path.
            xml_path: str, xml path.
            cover_thresh: float, cover rate about each subimage, default=0.2.
            iou_thresh: float, filter annotations which one's iou is smaller
                than thresh, default=1.0.

        Returns:
            subimages: a list of ImageWrapper.
            subannotations: a list of PascalXml
        """
        image = ImageWrapper().load(jpg_path)
        split_bboxes = self.get_crop_bboxes(width=image.width,
                                            height=image.height,
                                            min_side=min_side,
                                            max_side=max_side,
                                            cover_thresh=cover_thresh)
        subimages = image.crop_image(split_bboxes)

        xml_info = PascalXml().load(xml_path)
        subannotations = xml_info.crop_annotations(split_bboxes,
                                                   iou_thresh=iou_thresh)

        return subimages, subannotations

    def check_jpg_xml_match(self):
        """Check matching degree about xml files and jpeg files.
        """
        # arguemnts check
        assert os.path.exists(self.annotations.dir), self.annotations.dir
        assert os.path.exists(self.jpegimages.dir), self.jpegimages.dir

        # get name list
        xml_file_list = glob.glob(os.path.join(self.annotations.dir, '*.xml'))
        jpeg_file_list = glob.glob(os.path.join(self.jpegimages.dir, '*.jpg'))
        xml_name_list = [
            os.path.basename(path).split('.')[0] for path in xml_file_list
        ]
        jpeg_name_list = [
            os.path.basename(path).split('.')[0] for path in jpeg_file_list
        ]

        inter = list(set(xml_name_list).intersection(set(jpeg_name_list)))
        xml_diff = list(set(xml_name_list).difference(set(jpeg_name_list)))
        jpeg_diff = list(set(jpeg_name_list).difference(set(xml_name_list)))

        # print result and return matched list
        print('Find {} xml, {} jpg, matched {}.'.format(
            len(xml_file_list), len(jpeg_file_list), len(inter)))
        if len(xml_diff):
            print("Only have xml file: {}\n{}".format(len(xml_diff), xml_diff))
        if len(jpeg_diff):
            print("Only have jpg file: {}\n{}".format(len(jpeg_diff),
                                                      jpeg_diff))

        return inter



"""
Pascal Voc Tools

This package provide some tools using for
pascal voc format dataset and some usrful
functions.
"""

from .image_tools import ImageWrapper
from .xml_tools import DataSource
from .xml_tools import ImageSize
from .xml_tools import PascalXml
from .annotations_tools import Annotations
from .jpegimages_tools import JPEGImages
from .voc_tools import VOCTools
from .anchors_kmeans import AnchorsKMeans
from .darknet_config import DarknetConfig

__all__ = [
    'ImageWrapper', 'DataSource', 'ImageSize', 'PascalXml', 'JPEGImages',
    'Annotations', 'VOCTools', 'AnchorsKMeans', 'DarknetConfig'
]

name = 'pascal_voc_tools'



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
Calculating anchors using k-means for yolov3.
"""

import numpy as np
import os
import random

__DEBUG__ = False


def _width_and_height_iou(x, centroids):
    """
    Args:
        x: 某一个ground truth的w,h;
        centroids: anchor的w,h的集合[(w,h),(),...]，共k个;

    Returns:
        单个ground truth box与所有k个anchor box的IoU值集合;
    """
    IoUs = []
    w, h = x  # ground truth的w,h
    for centroid in centroids:
        c_w, c_h = centroid  # anchor的w,h
        if c_w >= w and c_h >= h:  # anchor包围ground truth
            iou = w * h / (c_w * c_h)
        elif c_w >= w and c_h <= h:  # anchor宽矮
            iou = w * c_h / (w * h + (c_w - w) * c_h)
        elif c_w <= w and c_h >= h:  # anchor瘦长
            iou = c_w * h / (w * h + c_w * (c_h - h))
        else:  # ground truth包围anchor
            # means both w,h are bigger than c_w and c_h respectively
            iou = (c_w * c_h) / (w * h)
        IoUs.append(iou)  # will become (k,) shape
    return np.array(IoUs)


def _avg_IOU(X, centroids):
    """
    Args:
        X: ground truth的w,h的集合[(w,h),(),...]
        centroids: anchor的w,h的集合[(w,h),(),...]，共k个

    Returns:
        centroids与ground truth 的平均iou
    """
    n, d = X.shape
    sum = 0.
    for i in range(X.shape[0]):
        # 返回一个ground truth与所有anchor的IoU中的最大值
        sum += max(_width_and_height_iou(X[i], centroids))
    return sum / n  # 对所有ground truth求平均


def _write_anchors_to_file(anchors, X, anchor_file):
    """

    Args:
        centroids: anchor的w,h的集合[(w,h),(),...]，共k个
        X: ground truth的w,h的集合[(w,h),(),...]
        anchor_file: anchor和平均IoU的输出路径

    Returns:
        anchors: saved anchors.
    """
    with open(anchor_file, 'w') as f:
        anchors_str = []
        for i in len(anchors):
            anchors_str.append('{:.2f},{:.2f}'.format(anchors[i, 0],
                                                      anchors[i, 1]))
        f.write(', '.join(anchors_str) + '\n')
        f.write('%f\n' % (_avg_IOU(X, anchors)))


def _kmeans(ground_truth_list, centroids):
    """以长宽的iou作为距离度量，以centroids为质心聚类

    Args:
        ground_truth_list: 所有ground truth的长宽
        centroids: 初始化的质心

    """
    ground_truth_number = ground_truth_list.shape[0]
    print("centroids.shape", centroids)
    anchor_number, anchor_dim = centroids.shape  # anchor的个数k以及w,h两维，dim默认等于2
    prev_assignments = np.ones(ground_truth_number) * \
        (-1)  # 对每个ground truth分配初始标签
    iterator = 0
    # 初始化每个ground truth对每个anchor的IoU
    old_all_distance = np.zeros((ground_truth_number, anchor_number))

    while True:
        all_distance = []
        iterator += 1
        for i in range(ground_truth_number):
            distance = 1 - \
                _width_and_height_iou(ground_truth_list[i], centroids)
            all_distance.append(distance)
        # 得到每个ground truth对每个anchor的IoU
        all_distance = np.array(all_distance)

        # 计算每次迭代和前一次IoU的变化值
        print("iter {}: dists = {}".format(
            iterator, np.sum(np.abs(old_all_distance - all_distance))),
              end='\r')

        # assign samples to centroids
        # 将每个ground truth分配给距离d最小的anchor序号
        assignments = np.argmin(all_distance, axis=1)

        # 如果前一次分配的结果和这次的结果相同，就输出anchor以及平均IoU
        if (assignments == prev_assignments).all():
            break

        # calculate new centroids
        # 初始化以便对每个簇的w,h求和
        centroid_sums = np.zeros((anchor_number, anchor_dim), np.float)
        # 将每个簇中的ground truth的w和h分别累加
        for i in range(ground_truth_number):
            centroid_sums[assignments[i]] += ground_truth_list[i]
        # 对簇中的w,h求平均
        for j in range(anchor_number):
            centroids[j] = centroid_sums[j] / (np.sum(assignments == j) + 1)

        prev_assignments = assignments.copy()
        old_all_distance = all_distance.copy()

    return centroids


class AnchorsKMeans():
    """对anchors聚类
    """
    def __init__(self, filelist):
        """like 2007_train.txt, data fromat must have labels

        Arguments:
            filelist: path to filelist;

        """
        self.filelist = filelist
        self.anchors = None
        self.ground_truth_list = None

    def save(self, output_dir):
        """Save current anchors to a file

        Args:
            output_dir: str, a dir to save anchors file.
        """
        # make sure the output_dir is exist.
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)

        anchor_file = os.path.join(output_dir, 'anchors%d.txt' %
                                   (len(self.anchors)))  # save path
        _write_anchors_to_file(self.anchors, self.ground_truth_list,
                               anchor_file)

    def calculate(self, num_clusters, yolo_input_shape, yolo_version='yolov3'):
        """
        Args:
            num_clusters: number of clusters;
            yolo_version: default='yolov3', yolov2 or yolov3;
            yolo_input_shape: input images shape，multiples of 32. etc. 416*416;

        Returns:
            anchors

        """
        # read filelist
        with open(self.filelist) as f:
            lines = [line.rstrip('\n') for line in f.readlines()]

        # get w and h for filelist
        annotation_dims = []
        for i, line in enumerate(lines):
            line = line.replace('JPEGImages', 'labels')
            line = line.replace('.jpg', '.txt')
            line = line.replace('.png', '.txt')
            print(str(i) + ': ' + line, end='\r')
            with open(line) as f2:
                for line in f2.readlines():
                    line = line.rstrip('\n')
                    w, h = line.split(' ')[3:]
                    annotation_dims.append((float(w), float(h)))
        print('\n')
        annotation_dims = np.array(annotation_dims)  # 保存所有ground truth框的(w,h)

        indices = [
            random.randrange(annotation_dims.shape[0])
            for i in range(num_clusters)
        ]
        centroids = annotation_dims[indices]
        centroids = _kmeans(annotation_dims, centroids)

        anchors = centroids.copy()
        if yolo_version == 'yolov2':
            for i in range(anchors.shape[0]):
                # yolo中对图片的缩放倍数为32倍，所以这里除以32，
                # 如果网络架构有改变，根据实际的缩放倍数来
                # 求出anchor相对于缩放32倍以后的特征图的实际大小（yolov2）
                anchors[i][0] *= yolo_input_shape / 32.
                anchors[i][1] *= yolo_input_shape / 32.
        elif yolo_version == 'yolov3':
            for i in range(anchors.shape[0]):
                # 求出yolov3相对于原图的实际大小
                anchors[i][0] *= yolo_input_shape
                anchors[i][1] *= yolo_input_shape
        else:
            print("the yolo version is not right!")
            exit(-1)

        widths = anchors[:, 0]
        sorted_indices = np.argsort(widths)

        print('Anchors = ', anchors[sorted_indices])
        self.anchors = anchors[sorted_indices]
        return self.anchors



# -*- coding:utf-8 -*-

import os
import numpy as np
import cv2
import tqdm
import random
from ._xml_parser import XmlParser


def pad_image(bg_image, pad_image, top_left):
    """pad a small image into a background image.

    Args:
        bg_image: ndarray, an image;
        pad_image: ndarray, an image;
        top_left: list, pad start point.
    Returns:
        masked image.
    Raise:
        AssertError: the pad size is not smaller than bg_image.
    """
    mask_image = bg_image.copy()
    # make sure the mask image is bigger than pad image.
    pad_image_height, pad_image_width = pad_image.shape[:2]
    assert mask_image.shape[0] >= top_left[0] + pad_image_height
    assert mask_image.shape[1] >= top_left[1] + pad_image_width

    mask_image[top_left[0]:top_left[0] + pad_image_height,
               top_left[1]:top_left[1] + pad_image_width] = pad_image
    return mask_image


def pad_object(bg_object, pad_obj, top_left):
    """
    """
    for obj in pad_obj:
        obj['bndbox']['xmin'] = str(int(obj['bndbox']['xmin']) + top_left[1])
        obj['bndbox']['ymin'] = str(int(obj['bndbox']['ymin']) + top_left[0])
        obj['bndbox']['xmax'] = str(int(obj['bndbox']['xmax']) + top_left[1])
        obj['bndbox']['ymax'] = str(int(obj['bndbox']['ymax']) + top_left[0])
    return bg_object + pad_obj


def pad2_horizon(image1, image2, image1_object, image2_object):
    """ Horizon concat two images.
    """
    bg_image = np.zeros(
        (max(image1.shape[0],
             image2.shape[0]), image1.shape[1] + image2.shape[1], 3),
        dtype=np.uint8)
    bg_image = pad_image(bg_image, image1, [0, 0])
    bg_image = pad_image(bg_image, image2, [0, image1.shape[1]])

    bg_annotation_object = pad_object(image1_object, image2_object,
                                      [0, image1.shape[1]])
    return bg_image, bg_annotation_object


def pad2_vertical(image1, image2, image1_object, image2_object):
    """ Vertical concat two images.
    """
    bg_image = np.zeros((image1.shape[0] + image2.shape[0],
                         max(image1.shape[1], image2.shape[1]), 3),
                        dtype=np.uint8)
    bg_image = pad_image(bg_image, image1, [0, 0])
    bg_image = pad_image(bg_image, image2, [image1.shape[0], 0])

    bg_annotation_object = pad_object(image1_object, image2_object,
                                      [image1.shape[0], 0])
    return bg_image, bg_annotation_object


def pad4(images_list, objects_list):
    """ 2*2 concat four images.
    """
    assert len(images_list) == 4
    assert len(objects_list) == 4

    bg_image1, bg_object1 = pad2_horizon(images_list[0], images_list[1],
                                         objects_list[0], objects_list[1])
    bg_image2, bg_object2 = pad2_horizon(images_list[2], images_list[3],
                                         objects_list[2], objects_list[3])
    bg_image, bg_object = pad2_vertical(bg_image1, bg_image2, bg_object1,
                                        bg_object2)
    return bg_image, bg_object


class PadDataset():
    """ Concat 2 or 4 images to one.

    Attributes:
        voc_root: str, PascalVOC foramt dataset;
        voc_save_root: str, save path;
        main_name: str, data set name.
    """
    def __init__(self, voc_root, voc_save_root, main_name='train'):
        self.voc_root = voc_root
        self.main_name = main_name
        self.jpeg_dir = os.path.join(voc_root, 'JPEGImages')
        self.annotation_dir = os.path.join(voc_root, 'Annotations')
        self.main_path = os.path.join(voc_root, 'ImageSets/Main',
                                      main_name + '.txt')

        assert os.path.exists(self.jpeg_dir)
        assert os.path.exists(self.annotation_dir)
        assert os.path.exists(self.main_path)

        self.save_jpeg_dir = os.path.join(voc_save_root, 'JPEGImages')
        self.save_annotation_dir = os.path.join(voc_save_root, 'Annotations')
        self.save_main_dir = os.path.join(voc_save_root, 'ImageSets/Main')
        if not os.path.exists(self.save_jpeg_dir):
            os.makedirs(self.save_jpeg_dir)
        if not os.path.exists(self.save_annotation_dir):
            os.makedirs(self.save_annotation_dir)
        if not os.path.exists(self.save_main_dir):
            os.makedirs(self.save_main_dir)

        with open(self.main_path) as f:
            names = f.read().strip().split('\n')
        self.ids = [n.strip() for n in names]
        print('Find {} data ids.'.format(len(self.ids)))

    def random_pad4(self):
        """ Concat 2*2 images.
        """
        random.shuffle(self.ids)
        save_ids = []
        for i in tqdm.tqdm(range(len(self.ids) // 4)):
            images = []
            annotation_objects = []
            for j in range(4):
                image_id = self.ids[i * 4 + j]
                image = cv2.imread(
                    os.path.join(self.jpeg_dir, image_id + '.jpg'))
                images.append(image)

                image_ann = XmlParser().load(
                    os.path.join(self.annotation_dir, image_id + '.xml'))
                if 'object' not in image_ann:
                    image_ann['object'] = []
                image_obj = image_ann['object']
                annotation_objects.append(image_obj)

            pad_image, pad_object = pad4(images, annotation_objects)
            save_id = '+'.join(self.ids[i * 4:i * 4 + 4])
            image_save_path = os.path.join(self.save_jpeg_dir,
                                           save_id + '.jpg')
            cv2.imwrite(image_save_path, pad_image)
            save_parser = XmlParser()
            save_parser.set_head(path=image_save_path,
                                 height=pad_image.shape[0],
                                 width=pad_image.shape[1])
            save_parser.template_parameters['object'] = pad_object
            save_parser.save(
                os.path.join(self.save_annotation_dir, save_id + '.xml'))
            save_ids.append(save_id)

        with open(os.path.join(self.save_main_dir, self.main_name + '.txt'),
                  'w') as f:
            f.write('\n'.join(save_ids))

    def random_pad2_horizon(self):
        """ Horizon concat 2 images.
        """
        random.shuffle(self.ids)
        save_ids = []
        for i in tqdm.tqdm(range(len(self.ids) // 2)):
            images = []
            annotation_objects = []
            for j in range(2):
                image_id = self.ids[i + j]
                image = cv2.imread(
                    os.path.join(self.jpeg_dir, image_id + '.jpg'))
                images.append(image)

                image_ann = XmlParser().load(
                    os.path.join(self.annotation_dir, image_id + '.xml'))
                if 'object' not in image_ann:
                    image_ann['object'] = []
                image_obj = image_ann['object']
                annotation_objects.append(image_obj)

            pad_image, pad_object = pad2_horizon(images[0], images[1],
                                                 annotation_objects[0],
                                                 annotation_objects[1])
            save_id = '+'.join(self.ids[i * 2:i * 2 + 2])
            image_save_path = os.path.join(self.save_jpeg_dir,
                                           save_id + '.jpg')
            cv2.imwrite(image_save_path, pad_image)
            save_parser = XmlParser()
            save_parser.set_head(path=image_save_path,
                                 height=pad_image.shape[0],
                                 width=pad_image.shape[1])
            save_parser.template_parameters['object'] = pad_object
            save_parser.save(
                os.path.join(self.save_annotation_dir, save_id + '.xml'))
            save_ids.append(save_id)

        with open(os.path.join(self.save_main_dir, self.main_name + '.txt'),
                  'w') as f:
            f.write('\n'.join(save_ids))

    def random_pad2_vertical(self):
        """ Vertical concat 2 images.
        """
        random.shuffle(self.ids)
        save_ids = []
        for i in tqdm.tqdm(range(len(self.ids) // 2)):
            images = []
            annotation_objects = []
            for j in range(2):
                image_id = self.ids[i + j]
                image = cv2.imread(
                    os.path.join(self.jpeg_dir, image_id + '.jpg'))
                images.append(image)

                image_ann = XmlParser().load(
                    os.path.join(self.annotation_dir, image_id + '.xml'))
                if 'object' not in image_ann:
                    image_ann['object'] = []
                image_obj = image_ann['object']
                annotation_objects.append(image_obj)

            pad_image, pad_object = pad2_vertical(images[0], images[1],
                                                  annotation_objects[0],
                                                  annotation_objects[1])
            save_id = '+'.join(self.ids[i * 2:i * 2 + 2])
            image_save_path = os.path.join(self.save_jpeg_dir,
                                           save_id + '.jpg')
            cv2.imwrite(image_save_path, pad_image)
            save_parser = XmlParser()
            save_parser.set_head(path=image_save_path,
                                 height=pad_image.shape[0],
                                 width=pad_image.shape[1])
            save_parser.template_parameters['object'] = pad_object
            save_parser.save(
                os.path.join(self.save_annotation_dir, save_id + '.xml'))
            save_ids.append(save_id)

        with open(os.path.join(self.save_main_dir, self.main_name + '.txt'),
                  'w') as f:
            f.write('\n'.join(save_ids))



import os
import unittest

from pascal_voc_tools import DataSource, ImageSize, PascalXml


class TestPascalXml(unittest.TestCase):
    save_path = './tests/test.xml'

    def tearDown(self):
        if os.path.exists(self.save_path):
            os.remove(self.save_path)

    def test_init(self):
        folder = './tests'
        filename = 'test.xml'
        path = './tests/test.jpg'
        source = DataSource('dataset')
        size = ImageSize(width=10, height=10, depth=3)
        segmented = 0
        xml = PascalXml(folder, filename, path, source, size, segmented)
        self.assertEqual(xml.folder, folder)
        self.assertEqual(xml.filename, filename)
        self.assertEqual(xml.path, path)
        self.assertEqual(xml.source, source)
        self.assertEqual(xml.size, size)
        self.assertEqual(xml.segmented, segmented)

    def test_save(self):

        save_data = "<annotation><folder></folder><filename></filename>" + \
            "<path></path><source><database>Unknown</database></source>" + \
            "<segmented>0</segmented><size><width>0</width>" + \
            "<height>0</height><depth>0</depth></size></annotation>"
        voc = PascalXml()
        voc.save(self.save_path)

        self.assertTrue(os.path.exists(self.save_path))

        with open(self.save_path) as f:
            data = f.read().strip()
            data = ''.join([line.strip() for line in data.split('\n')])
        self.assertEqual(data, save_data)



# -*- encoding: utf-8 -*-
"""
A wrapper for Annotations directory.
"""

import os
import glob
import csv
import logging
import tqdm
import numpy as np
from .xml_tools import PascalXml
import matplotlib.pyplot as plt

logger = logging.getLogger(__name__)


def split_list(val, split_num):
    max_val = max(val)
    min_val = min(val)
    x = list(np.arange(min_val, max_val,
                       (max_val - min_val) / split_num)) + [max_val]

    normalized_val = [(n - min_val) / (max_val - min_val) for n in val]
    normalized_val = [int(n * split_num) for n in normalized_val]
    y = [normalized_val.count(n) for n in range(split_num + 1)]
    return x, y


class Annotations():
    """ deal with annotations in PascalVOC format dataset.

    Attributes:
        ann_dir: str, the dir including xmls.
    Raises:
        AssertError: can not find path.
    """
    def __init__(self, ann_dir):
        self.dir = ann_dir
        self.ann_list = []

    def load(self, name_list=None):
        if name_list is None:
            self.ann_list = sorted(
                glob.glob(os.path.join(self.dir, '*.xml')))
        else:
            self.ann_list = [
                os.path.join(self.dir, '{}.xml'.format(name))
                for name in name_list
            ]
        logger.info('Find {} xml files.'.format(len(self.ann_list)))
        return self

    def count_category(self):
        name_dict = {}
        for xml_path in self.ann_list:
            xml_data = PascalXml().load(xml_path)
            xml_name_list = [obj.name for obj in xml_data.object]
            for name in xml_name_list:
                if name not in name_dict:
                    name_dict[name] = {'count': 0, 'included_file': []}
                name_dict[name]['count'] += 1
                file_name = os.path.basename(xml_path)
                if file_name not in name_dict[name]['included_file']:
                    name_dict[name]['included_file'].append(file_name)
        return name_dict

    def get_bbox_info(self):
        id_bbox_map = {}
        for xml_path in self.ann_list:
            xml_data = PascalXml().load(xml_path)
            size = [int(xml_data.size.heihgt), int(xml_data.size.width)]
            if 0 in size:
                print('Warrning: {} size error: {}'.format(xml_path, size))
                continue
            objects = xml_data.object
            id_bbox_map[os.path.basename(xml_path)] = objects
        return id_bbox_map

    def iou_analyse(self, save_dir='./', split_num=100):
        """ Draw iou distribution in dataset.

        Args:
            save_dir: str, image save dir;
            split_num: int, the number of data partitions
        """
        class_iou_map = {}
        for xml_path in self.ann_list:
            xml_data = PascalXml().load(xml_path)
            width = int(xml_data.size.width)
            height = int(xml_data.size.height)
            image_area = width * height

            for obj in xml_data.object:
                if obj.name not in class_iou_map:
                    class_iou_map[obj.name] = []

                xmin = int(obj.bndbox.xmin)
                ymin = int(obj.bndbox.ymin)
                xmax = int(obj.bndbox.xmax)
                ymax = int(obj.bndbox.ymax)
                roi_area = (xmax - xmin) * (ymax - ymin)
                class_iou_map[obj.name].append(roi_area / image_area)

        for key, val in class_iou_map.items():
            x, y = split_list(val, split_num)

            plt.figure(figsize=(8, 4))
            plt.plot(x, y, "bD-", linewidth=1)
            plt.xlabel("IOU(object area/image area)")
            plt.ylabel("Times")
            plt.title("The distribution of Objects' IOU in the dataset")
            plt.savefig(
                os.path.join(save_dir, "IOU_distribution-{}.jpg".format(key)))
        return class_iou_map

    def height_analyse(self, save_dir='./', split_num=100):
        """ Draw height ratio distribution in dataset.

        Args:
            save_dir: str, image save dir;
            split_num: int, the number of data partitions
        """
        class_iou_map = {}
        for xml_path in self.ann_list:
            xml_data = PascalXml().load(xml_path)
            height = int(xml_data.size.height)

            for obj in xml_data.object:
                if obj.name not in class_iou_map:
                    class_iou_map[obj.name] = []

                ymin = int(obj.bndbox.ymin)
                ymax = int(obj.bndbox.ymax)
                class_iou_map[obj.name].append((ymax - ymin) / height)

        for key, val in class_iou_map.items():
            x, y = split_list(val, split_num)

            plt.figure(figsize=(8, 4))
            plt.plot(x, y, "bD-", linewidth=1)
            plt.xlabel("Height ratio(object/image) percent/%")
            plt.ylabel("Times")
            plt.title("The distribution of Objects' height in the dataset")
            plt.savefig(
                os.path.join(save_dir,
                             "Height_distribution-{}.jpg".format(key)))
        return class_iou_map

    def width_analyse(self, save_dir='./', split_num=100):
        """ Draw width ratio distribution in dataset.

        Args:
            save_dir: str, image save dir;
            split_num: int, the number of data partitions
        """
        class_iou_map = {}
        for xml_path in self.ann_list:
            xml_data = PascalXml().load(xml_path)
            width = int(xml_data.size.width)

            for obj in xml_data.object:
                if obj.name not in class_iou_map:
                    class_iou_map[obj.name] = []

                xmin = int(obj.bndbox.xmin)
                xmax = int(obj.bndbox.xmax)
                width_ratio = (xmax - xmin) / width

                class_iou_map[obj.name].append(width_ratio)

        for key, val in class_iou_map.items():
            x, y = split_list(val, split_num)

            plt.figure(figsize=(8, 4))
            plt.plot(x, y, "bD-", linewidth=1)
            plt.xlabel("Width ratio(object/image) percent/%")
            plt.ylabel("Times")
            plt.title("The distribution of Objects' width in the dataset")
            plt.savefig(
                os.path.join(save_dir,
                             "Width_distribution-{}.jpg".format(key)))
        return class_iou_map

    def convert2csv(self, save_path, category_list=[]):
        """find all bndbox info and save to csv file. If category_list exist, only save
        the categories in list.

        Arguments:
            save_path: a csv file path.
            category_list: a list for keep category to save.

        Raises:
            AssertionError: the save path not a csv file path.
        """
        header = [
            'file_name', 'width', 'height', 'category', 'xmin', 'ymin', 'xmax',
            'ymax'
        ]

        xmls_list = sorted(glob.glob(os.path.join(self.dir, '*.xml')))
        logger.info(f"Find {len(xmls_list)} xmls")

        assert '.csv' == save_path[
            -4:], "Make sure the save path is a new csv file path"
        f = open(save_path, 'w')
        f_csv = csv.writer(f)
        f_csv.writerow(header)

        # bbox_info_list = []
        for xml_path in tqdm.tqdm(xmls_list):
            # print(xml_path)
            bbox_list = PascalXml().load(xml_path).convert2csv(category_list)
            f_csv.writerows(bbox_list)

        f.close()

        return self



import os
import glob
import numpy as np
import cv2

from _xml_parser import XmlParser


class FalsePositiveArgument:
    """
    """
    def __init__(self, fp_image_list, voc_root_dir, save_dir):
        """
        """
        self.fp_image_list = fp_image_list
        self.voc_root_dir = voc_root_dir
        self.save_dir = save_dir

        self.images_dir = os.path.join(voc_root_dir, 'JPEGImages')
        self.xmls_dir = os.path.join(voc_root_dir, 'Annotations')

    def image_augment(self, image):
        """增强贴图

        Args:
            image: ndarray, an opencv image;
        Returns:
            an augmented image.
        """
        image_augmented = image
        return image_augmented

    def load_backgroud_images(self):
        """
        """
        images = glob.glob(os.path.join(self.images_dir, '*.jpg'))
        return images

    def load_annotation(self, image_id):
        """倒入一个xml数据

        Arguments:
            image_id: str, an image id;
        Returns:
            the xml data about input image id.
        Raises:
            AssertError: can not find corresponding xml file.
        """
        xml_path = os.path.join(self.xmls_dir, image_id + '.xml')
        assert os.path.exists(xml_path), xml_path

        parser = XmlParser()
        xml_data = parser.load(xml_path)
        return xml_data

    def get_used_mask(self, image, xml_data):
        """根据xml数据，产生一个mask，标注有目标的位置

        Arguments:
            image: ndarray, an opencv image;
            xml_data: dict, xml data;
        Returns:
            image_mask
        """
        mask = np.zeros((image.shape[0], image.shape[1]))

        for obj in xml_data['object']:
            xmin = int(obj['xmin'])
            ymin = int(obj['ymin'])
            xmax = int(obj['xmax'])
            ymax = int(obj['ymax'])

            mask[ymin:ymax, xmin:xmax] = 1
        return mask

    def ramdom_pad(self, image_path, fp_image, try_time=10):
        """try to pad fp_image to image.
        """
        assert os.path.exists(image_path), image_path

        image = cv2.imread(image_path)
        image_id = os.path.basename(image_path).split('.')[0]
        xml_data = self.load_annotation(image_id)
        mask = self.get_used_mask(image, xml_data)

        for try_index in range(try_time):
            # TODO: 随机选择一个mask为0的位置，判断以该位置为左上角点是否可行
            pass



# coding:utf-8
import glob
import logging
import os

import cv2
import numpy as np
from PIL import Image

logger = logging.getLogger(__name__)


class ImageWrapper(object):
    """Some image tools.
    """
    def __init__(self):
        self.path = None
        self.data = None

        self.width = None
        self.height = None
        self.depth = None

    def load(self, image_path: str):
        """load image data.

        Arguments:
            image_path: image file path.
        """
        assert os.path.exists(image_path), image_path
        self.data = cv2.imread(image_path)
        self.path = image_path
        self.height = self.data.shape[0]
        self.width = self.data.shape[1]
        self.depth = self.data.shape[2] if len(self.data.shape) == 3 else 1
        return self

    def resize_by_rate(self, rate: float):
        """Resize image.

        Arguments:
            rate: image resize rate.
        """
        self.data = cv2.resize(self.data, None, fx=rate, fy=rate)
        self.height = self.data.shape[0]
        self.width = self.data.shape[1]
        return self

    def resize_letter_box(self, width: int, height: int):
        """using letter box to resize image data.

        Arguments:
            width: new image width.
            height: new image height.

        Returns:
            a flot of resize rate.
        """
        if self.depth == 1:
            mask_image = np.zeros((height, width), dtype=np.uint8)
        else:
            mask_image = np.zeros((height, width, self.depth), dtype=np.uint8)

        rate = min(float(width) / self.width, float(height) / self.height)
        new_width = int(self.width * rate)
        new_height = int(self.height * rate)

        horizion_bias = int((width - new_width) / 2)
        vertical_bias = int((height - new_height) / 2)

        resized_image = cv2.resize(self.data, (new_width, new_height))
        mask_image[vertical_bias:vertical_bias + new_height,
                   horizion_bias:horizion_bias + new_width] = resized_image

        self.data = mask_image
        self.height = new_height
        self.width = new_width
        return rate, (vertical_bias, horizion_bias)

    def save(self, save_path):
        """save image data to a file.

        Arguments:
            save_path: a image file path.
        """
        cv2.imwrite(save_path, self.data)
        return self

    def crop_image(self, split_bboxes):
        """Split an image to some subimages.

        Arguments:
            split_bboxes: list, like[[xmin, ymin, xmax, ymax], ]

        Returns:
            images: list, all subimages.
        """
        subimages = []
        for bbox in split_bboxes:
            sub_image = ImageWrapper()
            sub_image.data = self.data[bbox[1]:bbox[3], bbox[0]:bbox[2]]
            sub_image.width = bbox[2] - bbox[0]
            sub_image.height = bbox[3] - bbox[1]
            sub_image.depth = self.depth
            subimages.append(sub_image)
        return subimages


def verify_image(jpeg_path):
    """Verify image format.
    Args:
        jpeg_path: str, the image path.

    Returns:
        result: bool, if image ok, True will return.
    """
    assert os.path.exists(jpeg_path), jpeg_path
    try:
        Image.open(jpeg_path).verify()
    except Exception as err:
        logger.exception(err)
        return False
    return True


def image_convert_2_jpg(images_dir):
    """
    Change image format from others to jpg.
    Args:
        jpeg_dir: str, the dir only have image.
    """
    image_path_list = glob.glob(os.path.join(images_dir, '*.*'))
    for jpeg_path in image_path_list:
        suffix = os.path.basename(jpeg_path).split('.')[-1]
        if suffix != 'jpg' and verify_image(jpeg_path):
            image = Image.open(jpeg_path)
            save_path = jpeg_path.replace('.{}'.format(suffix), '.jpg')
            image.save(save_path)
            os.remove(jpeg_path)
            print('Image from {} to {}'.format(jpeg_path, save_path))

    return True



import os
import unittest
import shutil
from pascal_voc_tools import VOCTools


class TestVOCTools(unittest.TestCase):
    save_dir = os.path.join(os.path.dirname(__file__), 'gen_devkit')

    def setUp(self):
        os.makedirs(self.save_dir)

    def tearDown(self):
        shutil.rmtree(self.save_dir)

    def test_get_year(self):
        voc_dir = os.path.join(self.save_dir, 'VOC2007')

        voc = VOCTools(voc_dir)
        self.assertTrue(voc.year == '2007')

    def test_gen_format_dir(self):
        voc_dir = os.path.join(self.save_dir, 'VOC2007')
        voc = VOCTools(voc_dir).gen_format_dir()

        # dir exists assert
        save_ann_dir = os.path.join(voc_dir, 'Annotations')
        save_jpeg_dir = os.path.join(voc_dir, 'JPEGImages')
        save_main_dir = os.path.join(voc_dir, 'ImageSets/Main')
        self.assertTrue(os.path.exists(voc_dir))
        self.assertTrue(os.path.exists(save_ann_dir))
        self.assertTrue(os.path.exists(save_jpeg_dir))
        self.assertTrue(os.path.exists(save_main_dir))
